package Analysis;
import java.util.ArrayList;
import java.util.List;
//import java.util.Collections;
//import java.util.List;

import Common.Constants;
import Common.Enums;
import Common.Task;
import Common.Taskset;
import Common.PartitionIterable;
//import Common.Utils;

public class PartitionedMembergExhaustive 
{
	private ArrayList<Task> taskSet; 
	private int numberOfCores; 
	private Enums.ActionResult schedulable;	// schedulability of input task set
	private Part [] parts; 					// parts: one per core
	private int memBW;						// system memory BW	
	private int memBWLeft;					// memory BW not allocated yet
	private int analOption;					// analysis option: transient (0) vs steady (1)
	private Part [] subsets;				// NEW: all subsets of tasks in the task-set
	
	
	private class Part {					// set of tasks assigned to a core
		private PartitionedMembergExhaustive heuristic;
		private int core;					// which core
		private ArrayList<Task> taskSet;	// which tasks
		
		private Enums.ActionResult schedulable;	// schedulability of this partition
		private int memBudget;			// memory budget assigned to this partition/core
		private int id;					// NEW: memory id, generated using a bitmask
			
		Part(int c, PartitionedMembergExhaustive h) {
			heuristic = h;
			core = c;
			taskSet = new ArrayList<Task>();
			schedulable = Enums.ActionResult.Failure;
			memBudget = 0;
		}
				
		void genId() {
			id = 0;
			for (Task t: taskSet ) {
				id += 1<<t.getTaskNo();
			}
		}
		Common.Enums.ActionResult analyze(int memBWLeft) {
			
			if(!taskSet.isEmpty() && memBWLeft == 0)
				return Common.Enums.ActionResult.Failure;
			if(taskSet.isEmpty())
				return Common.Enums.ActionResult.Success;
			
			MembergAnalysis analyzer = new MembergAnalysis(taskSet, memBWLeft, 
															heuristic.numberOfCores, 
															heuristic.analOption);
			schedulable = analyzer.GetSchedulabilityStatus();
			if(schedulable == Common.Enums.ActionResult.Success) {
				memBudget = memBWLeft;
				minimizeMemBudget();	
			}
			return schedulable;
		}
		
		private void minimizeMemBudget () // assumes that the task-set is feasible with current_MemoryBandwidth
		{
			if(taskSet.isEmpty())
				return;

			int t = 0; 
			int first = 0, second = memBudget; 
			int diff = Constants.MemoryBandwidthQuantization; 
			Enums.ActionResult status = null; 

			while(second - first > diff)
			{
				t = (int)(Math.floor(((double)second + (double)first)/2)); // (first + second)/2 wouldbe enough
				status = new MembergAnalysis(taskSet, t, numberOfCores,
											 heuristic.analOption).GetSchedulabilityStatus(); 
				if( status == Enums.ActionResult.Success)
				{
					second = t;
				}
				else 
				{
					first = t; 
				}
			}
			memBudget = second;
		}
		
		void display() {
			System.err.println("\n*** Set of tasks of part (core) " + core + " ********\n");
			for(Task task: taskSet) {
				task.Print();
			}
			if( schedulable == Enums.ActionResult.Success )
				System.err.println("\tPart with memory budget: " + memBudget);
			else
				System.err.println("\tPart is not schedulable with: " + heuristic.memBWLeft);
		}
	} // class Part
	
	void genSubset(int n, ArrayList<Task>taskSet) {
		subsets[n] = new Part(-1, this);	// -1 means that this is not mapped to any core
		for(int i = 0, m = n; i < taskSet.size() && n > 0; i++, m /= 2) {
			if( m % 2 == 1 ) 
				subsets[n].taskSet.add(taskSet.get(i));
		}
	}

	public PartitionedMembergExhaustive(Taskset ts, int m)
	{
		numberOfCores = m;
	
		taskSet = new ArrayList<Task>(ts.GetSize()); 
		Task task = ts.GetHeadOfTheList();
		
//		System.err.println("\n\n*********** Exhaustive Search (Taskset " + task.GetID()/32 + ") ******************");

		for(int taskNo = 0; task!=null; task = task.GetNextTask(), taskNo++)
		{
			// NEW: assign a taskNo to each task of the task-set
			//      this is used to generate a partition id
			task.setTaskNo(taskNo);
			taskSet.add(task); 
			//System.out.println("Added task " + taskNo);
		}
		
		// NEW: generate all subsets of tasks in the task-set
		//       and analyze the schedulability of each of them
		subsets = new Part[1<<taskSet.size()];
		
		memBW = Common.Constants.totalMemoryBudget;
		memBWLeft = memBW;
		
		this.analOption = 0;	// transient analysis
		
		for(int i = 0; i < (1<<taskSet.size()); i++) {
			genSubset(i, taskSet);
			subsets[i].analyze(memBW);
			if( subsets[i].schedulable == Enums.ActionResult.Failure )
				subsets[i].memBudget = 2*memBW;
		}
		
//		System.out.println("\n\n###### Listing all subsets ############\n");
//		
//		for(int i = 0; i < subsets.length; i++) {
//			System.out.println("#########Subset " + i + " ################" );
//			subsets[i].display();
//		}
		
		/*** No need to sort task-set, as we will do an exhaustive search, if necessary ***/
		
		schedulable = Enums.ActionResult.Failure;
		
		parts = new Part[numberOfCores];
		
		mapAndTest3();
 	
		// displayCoreTasksets();

	}
	
	private void mapAndTest3() {	
		
		int partNo = 0;

		
		// for each partition
		for(int n = 1; n <= numberOfCores; n++) {
            for (List<List<Task>> partition : 
                new PartitionIterable<>(taskSet,n)) {
        		int memBW = 0;			// NEW
            	
//				System.err.println("\n****** Partition " + partNo + " ******");
				
				generateParts(partition);
				
				for(int i = 0; i < partition.size(); i++) {
					// NEW: compute part id
					parts[i].genId();
					// NEW: add demand
					memBW += subsets[parts[i].id].memBudget;
				}

				if( memBW <= Common.Constants.totalMemoryBudget )
					schedulable = Common.Enums.ActionResult.Success;
				else
					schedulable = Common.Enums.ActionResult.Failure;
				
//				for(int i = 0; i < partition.size(); i++) {
//					parts[i].display();
//				}
				
				if( schedulable == Common.Enums.ActionResult.Success ) {
//					System.err.println("\n ### Partition " + partNo + " is schedulable :) ###");
					return;
				}
				partNo++;
            }
		}
//		System.err.println("\n ### NO partition schedulable :( ###");
		// checked all the partitions: failure	
	}
	
	private void mapAndTest2() {	
		
		int partNo = 0;
		
		// for each partition
		for(int n = 1; n <= numberOfCores; n++) {
            for (List<List<Task>> partition : 
                new PartitionIterable<>(taskSet,n)) {
				memBWLeft = memBW;
				
				System.err.println("\n****** Partition " + partNo + " ******");
				
				generateParts(partition);
				
				for(int i = 0; i < partition.size(); i++) {
					if( parts[i].analyze(memBWLeft) == Common.Enums.ActionResult.Success) {
						memBWLeft -= parts[i].memBudget;
						schedulable = Common.Enums.ActionResult.Success;
					} else {
						schedulable = Common.Enums.ActionResult.Failure;
						break;
					}
				}
				
				for(int i = 0; i < partition.size(); i++) {
					parts[i].display();
				}
				
				if( schedulable == Common.Enums.ActionResult.Success ) {
					System.err.println("\n ### Partition " + partNo + " is schedulable :) ###");
					return;
				}
				partNo++;
            }
		}
		System.err.println("\n ### NO partition schedulable :( ###");
		// checked all the partitions: failure	
	}
	
	private void mapAndTest() {	
		
		// for each partition
		for(int i = 0; i < 128; i++ ) {
			memBWLeft = memBW;
			if(numberOfCores != 2
					|| taskSet.size() != 8 ) {
				System.err.println("FAILURE: analysis supports only 2 cores and 8 tasks");
				System.exit(1);
			}
			
			System.err.println("\n****** Partition " + i + " ******");
			
			generatePartition(i);
//			for(int j = 0; j < numberOfCores; j++)
//				parts[j].display();
			if( parts[0].analyze(memBWLeft) == Common.Enums.ActionResult.Success ) {
				memBWLeft -= parts[0].memBudget;
				if( parts[1].analyze(memBWLeft) == Common.Enums.ActionResult.Success) {
					schedulable = Common.Enums.ActionResult.Success;
					System.err.println("\n ### Partition " + i + " is schedulable :) ###");
					return;
				}				
			}
		}
		System.err.println("\n ### NO partition schedulable :( ###");
		// checked all the partitions: failure	
	}
	
	// partitions the task set in two parts
	void generatePartition(int i) {	
		
		for(int j = 0; j < numberOfCores; j++)
			parts[j] = new Part(j,this);

		for(Task task: taskSet) {
			parts[i%2].taskSet.add(task);
			i /= 2;
		}
		
	}
	
	void generateParts(List<List<Task>> partition) {
		for(int j = 0; j < numberOfCores; j++)
			parts[j] = new Part(j,this);
		int i = 0;
		for(List<Task> part: partition ) {
			for(Task t: part) {	// it is unfortunate that we are using ArrayList<Task> rather than List<Task> 
				parts[i].taskSet.add(t);
			}
			i++;
		}
	}
	
	// XXX generates the partitions from the list<list<Tasks>>
	//     basically, the list<Tasks> is the set of tasks in each part (also called block, elsewhere)
	//     note that in some cases, the number of parts is less than the number of cores

	void displayPart(int [] p, String str) {	
		System.err.print(str + ": (");
		for(int i = 0; i < p.length ; i++) 
			System.err.print( p[i] + (i == p.length-1 ? ")\n" : ", "));
		System.err.flush();
	}
	
	
	// Recursively generates the partitions
	// 		c - core whose first task we need to choose next
	//  	q - array with the first task of each core of the current partition
	// We start with c = 1, the first core:
	//	by choosing the first task of that core, we determine the part on core 0
	//  after creating that part, we check its schedulability, and 
	//  if schedulable, we minimize the bandwidth assigned to that partition
	//  	and move to the next core
	//  if not schedulable, we prune some branches:
	//      e.g. if makes no sense to try larger values of d, as the taskset would no be
	//        schedulable anyway
	// We terminate the recursion when c = m
	//  here we just check the schedulability of the last core (m-1),
	//   because the last task is always the last task of that core 
	//
	//	Inputs:
	//		c - core being assigned
	//		q - first tasks of current partition (only for cores up to c)
	//
	
	void recGenParts(int c, int []q) {	
		// we try the different partitions, depending on their "distance" to the pivot partition
		// thus to avoid repeating the analysis, 
		// we keep track of the partitions that were already tried
		boolean [] tried = new boolean[1];

		// apparently this is already initialized with false, but just to be sure
		for(int i = 0; i < tried.length; i++)
			tried[i] = false;
		
		for( int d = 0; d <= 10; d++ ) {	// try each task between pivot-delta and pivot+delta
			for(int i = -d; i <= d; i += d == 0 ? 1 : 2*d ) { // first, try the tasks closer to the pivot
				// for d = 0, we have only to test one case, hence the weird exit condition
				// note that selecting the first task of core c, 
				//  determines the tasks in the part of the previous core (c-1)
				// but for the last core, the last task is always the last task of that core's part
				if( c < numberOfCores ) { 
		

					// create Part for this core:
					parts[c] = new Part(c, this);
				}
				// add tasks to the part mapped onto the previous core
				// note that for the last core with tasks, the last task is always taskset.size()-1

				
				// check schedulability of previous core's part
				if( parts[c-1].analyze(memBWLeft) == Common.Enums.ActionResult.Failure ) {
					
//					System.err.print("Core " + (c-1) + " is not schedulable: ");
//					System.err.print(q[0]);
//					for(int k = 1; k <= c && k < numberOfCores; k++) {
//						System.err.print(" , " + q[k]);
//					}
//					System.err.println();
					
					if( c < numberOfCores ) { 

						// not schedulable: prune the partitions that have a larger number of tasks
						for(int j =  i + 1; j < 2  + 1; j++ ) {
							
//							System.err.print("\tPrunning partition: ");
//							for(int k = 0; k < c; k++) {
//								System.err.print(q[k] + " , ");
//							}
//							//System.err.println(q[c]+j-delta-1);
//							System.err.println(pivot[c]+j-1);
							
							tried[j] = true; // XXX: we may do some double work here, but OK
						}
						continue; // try next partition, if any, with lower number of tasks
					} else
						return; // there is nothing else to try in this call, undo recursion
				}
				
//				parts[c-1].display();
//				System.err.print("Core " + (c-1) + " is schedulable: ");
//				System.err.print(q[0]);
//				for(int k = 1; k < c+1 && k < numberOfCores; k++) {
//					System.err.print(" , " + q[k]);
//				}
//				System.err.println();
				
				// The part of the previous core is schedulable
				memBWLeft -= parts[c-1].memBudget;			
				
				if( c < numberOfCores  					
						&& q[c] < taskSet.size()) { // this is not the last core with tasks assigned
					// select first task of next core
					recGenParts(c+1,q);
					if( schedulable == Enums.ActionResult.Success) // found a schedulable partition :)
						return; // so we are done
					else 
						memBWLeft += parts[c-1].memBudget;				
				} else { // the previous core was the last one with tasks: found a schedulable partition :)
					schedulable = Enums.ActionResult.Success;
					return; // undo recursion
				}
			} // inner loop
		} // outer loop		
	}
			

	
	public Enums.ActionResult GetSchedulabilityStatus()
	{
		return schedulable; 
	}

}
